"""
@file
@brief Exports an ONNX graph in a way it can we created again
with a python script. It relies on :epkg:`jinja2` and :epkg:`autopep8`.

.. versionadded:: 0.7
"""
import textwrap
import numpy
import onnx
from onnx.helper import printable_graph, make_node
from onnx import numpy_helper, ModelProto
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from .onnx2py_helper import (
    _var_as_dict, guess_proto_dtype, guess_proto_dtype_name,
    get_tensor_shape, get_tensor_elem_type)
from .onnx_export_templates import (
    get_onnx_template, get_tf2onnx_template, get_numpy_template,
    get_xop_template, get_cpp_template, get_python_template)
from .exports.numpy_helper import make_numpy_code
from .exports.tf2onnx_helper import make_tf2onnx_code


_keywords = {
    'False', 'await', 'else', 'import', 'pass',
    'None', 'break', 'except', 'in', 'raise',
    'True', 'class', 'finally', 'is', 'return',
    'and', 'continue', 'for', 'lambda', 'try',
    'as', 'def', 'from', 'nonlocal', 'while',
    'assert', 'del', 'global', 'not', 'with',
    'async', 'elif', 'if', 'or', 'yield'}


def _rename_var(var, empty='None'):
    if var in _keywords:
        return 'r_' + var
    if var == '':
        return empty
    return var


def select_attribute(ens, att, sort=False, unique=False, skip=None):
    """
    Returns the list of the same attribute.
    `[el.att for el in ens]`.

    :param ens: list
    :param att: attribute name
    :param sort: sort the array
    :param unique: returns the unique values
    :param skip: to skip some names
    :return: something like `[el.att for el in ens]`
    """
    if len(ens) == 0:
        return []
    if isinstance(ens[0], dict):
        atts = [el[att] for el in ens]
    else:
        atts = [getattr(el, att) for el in ens]
    if unique:
        atts = list(set(atts))
    if sort:
        atts.sort()
    if skip is None:
        return atts
    return [a for a in atts if a not in skip]


def _nodes(graph, rename_name, used, output_names, use_onnx_tensor,
           templates, verbose, opset, rename, autopep_options, name,
           subgraphs, unique_operators, opsets=None):
    if opsets is None:
        raise ValueError(  # pragma: no cover
            "opsets cannot be None.")
    if unique_operators is not None:
        from ..npy.xop import loadop
    nodes = []
    for node in graph.node:
        if (unique_operators is not None and
                node.domain in ('', 'ai.onnx.ml')):
            clname = loadop((node.domain, node.op_type))
            unique_operators.add(
                (node.domain, node.op_type, clname.__name__))
        for i_raw_name in node.input:
            if len(i_raw_name) == 0:
                i = 'None'
            else:
                i = rename_name(i_raw_name, out=False)
                if i not in used:
                    used[i] = []
                used[i].append(node)
        attributes = []
        for at in node.attribute:
            temp = _var_as_dict(at)
            value = temp['value']
            if node.op_type in {'Scan', 'Loop'} and at.name == 'body':
                if "{{ inputs[0][0] }}" in str(templates):
                    attributes.append((at.name, at.g))
                    continue
                fname = "_create_" + node.op_type + "_" + node.name + "_body"
                body = export_template(
                    value, templates, opset=opset, verbose=verbose,
                    name=name, rename=rename,
                    use_onnx_tensor=use_onnx_tensor,
                    autopep_options=autopep_options,
                    function_name=fname, opsets=opsets)
                subgraphs.append(
                    (body, node.op_type + "_" + node.name + "_body"))
                attributes.append((at.name, fname + "()"))
                continue
            if node.op_type == 'If' and at.name in {'then_branch', 'else_branch'}:
                if "{{ inputs[0][0] }}" in str(templates):
                    attributes.append((at.name, at.g))
                    continue
                fname = "_create_if_" + node.name + "_" + at.name
                body = export_template(
                    value, templates, opset=opset, verbose=verbose,
                    name=name, rename=rename,
                    use_onnx_tensor=use_onnx_tensor,
                    autopep_options=autopep_options,
                    function_name=fname, opsets=opsets)
                subgraphs.append((body, "if_" + node.name + "_" + at.name))
                attributes.append((at.name, fname + "()"))
                continue
            if use_onnx_tensor:
                if node.op_type == 'Cast' and at.name == 'to':
                    attributes.append(
                        (at.name, guess_proto_dtype_name(int(value))))
                    continue
            if isinstance(value, str):
                attributes.append((at.name, f"{value!r}"))
            else:
                if isinstance(value, numpy.ndarray):
                    if use_onnx_tensor and at.name == 'value':
                        onnx_dtype = guess_proto_dtype_name(
                            guess_proto_dtype(value.dtype))
                        value = (
                            'make_tensor("value", %s, dims=%r, vals=%r)'
                            '' % (onnx_dtype, list(value.shape),
                                  value.tolist()))
                        attributes.append((at.name, value))
                    else:
                        attributes.append((at.name, repr(value.tolist())))
                else:
                    attributes.append((at.name, repr(value)))

        attributes_str = ", ".join(f"{k}={v}" for k, v in attributes)
        d = dict(name=node.name, op_type=node.op_type,
                 domain=node.domain, onnx_node=node,
                 inputs=[rename_name(n, out=False)
                         for n in node.input if len(n) > 0],
                 outputs=[rename_name(n, out=True) for n in node.output],
                 output_names=[rename_name(n, out=True) for n in node.output
                               if n in output_names],
                 attributes=attributes, attributes_str=attributes_str)
        nodes.append(d)
    return nodes


def _xop_make_node_name(domain, name):
    from ..npy.xop import _domain_to_class_name
    class_name = "Onnx" + _domain_to_class_name(domain) + name
    return class_name


def _python_make_node_name(domain, version, name, node=False):
    if node:
        if version is None:
            version = 1
        if not isinstance(version, int):
            raise TypeError(  # pragma: no cover
                "version must be an integer not %r for domain=%r and name=%r." % (
                    version, domain, name))
        if domain == '':
            return "opset%d.%s" % (version, name)
        return "%s%d.%s" % (domain.replace(".", "_"), version, name)
    return name


def _python_make_node_graph(graph, opsets, indent=0, output_names=None):
    """
    Translates a GraphProto into python.
    """
    code = []
    sindent = '    ' * indent
    for init in graph.initializer:
        node = make_node('Constant', [], [_rename_var(init.name)], value=init)
        code.append(_python_make_node(node, opsets, indent=indent))
    if len(graph.sparse_initializer) > 0:
        raise NotImplementedError(  # pragma: no cover
            "Unable to convert sparse_initilizer into python.")
    for node in graph.node:
        code.append(_python_make_node(node, opsets, indent=indent))
    if output_names is not None:
        for fr, to in zip(graph.output, output_names):
            code.append(f"{sindent}{_rename_var(to)} = {_rename_var(fr.name)}")
    return "\n".join(code)


def _python_make_node_make_attribute_str(node):
    attributes = []
    for at in node.attribute:
        temp = _var_as_dict(at)
        value = temp['value']
        if isinstance(value, str):
            attributes.append((at.name, f"{value.decode('utf-8')!r}"))
            continue
        if isinstance(value, numpy.ndarray):
            if at.name == 'value':
                onnx_dtype = guess_proto_dtype_name(
                    guess_proto_dtype(value.dtype))
                value = (
                    'make_tensor("value", %s, dims=%r, vals=%r)'
                    '' % (onnx_dtype, list(value.shape),
                          value.ravel().tolist()))
                attributes.append((at.name, value))
                continue
            attributes.append((at.name, repr(value.tolist())))
            continue
        attributes.append((at.name, repr(value)))

    return ", ".join(f"{k}={v}" for k, v in attributes)


def _python_make_node_if(node, opsets, indent=0):
    """
    Translates a node If into python.
    """
    sindent = '    ' * indent
    code = [f"{sindent}if {node.input[0]}:"]
    if len(node.attribute) != 2:
        raise RuntimeError(  # pragma: no cover
            f"Node {node.op_type!r} expected two attributes not {len(node.attribute)}.")
    atts = node.attribute
    if atts[0].name == 'else_branch':
        else_branch, then_branch = atts[0].g, atts[1].g
    else:
        else_branch, then_branch = atts[1].g, atts[0].g
    code.append(_python_make_node_graph(
        then_branch, opsets, indent=indent + 1,
        output_names=node.output))
    code.append(f"{sindent}else:")
    code.append(_python_make_node_graph(
        else_branch, opsets, indent=indent + 1,
        output_names=node.output))
    return "\n".join(code)


def _python_make_node_loop(node, opsets, indent=0):
    """
    Translates a node Loop into python.
    """
    raise NotImplementedError()  # pragma: no cover


def _python_make_node_scan(node, opsets, indent=0):
    """
    Translates a node Scan into python.
    """
    raise NotImplementedError()  # pragma: no cover


def _python_make_node(onnx_node, opsets, indent=0):
    if isinstance(onnx_node, dict):
        node = onnx_node['onnx_node']
    else:
        node = onnx_node
    version = opsets[node.domain]
    if node.op_type in {'If', 'Loop', 'Scan'}:
        # If, Loop, Scan
        if node.op_type == 'If':
            return _python_make_node_if(node, opsets, indent=indent)
        if node.op_type == 'Loop':
            return _python_make_node_loop(node, opsets, indent=indent)
        if node.op_type == 'Scan':
            return _python_make_node_scan(node, opsets, indent=indent)
        raise RuntimeError(  # pragma: no cover
            f"Unable to export node type {node.op_type!r} into python.")
        # pragma: no cover
    if any(map(lambda att: hasattr(att, 'g') and att.g and att.g.ByteSize() > 0,
               node.attribute)):
        raise RuntimeError(  # pragma: no cover
            f"Unable to export node type {node.op_type!r} into python.")
    ops = {'Add': '+', 'Sub': '-', 'Mul': '*', 'MatMul': '@',
           'Div': '/', 'Pow': '**',
           'And': '&', 'Or': '|', 'Greater': '>', 'Equal': '==',
           'Lesser': '<', 'GreaterOrEqual': '>=', 'LessOrEqual': '<='}
    sindent = "    " * indent
    if node.op_type in ops:
        return "%s%s = %s" % (sindent, _rename_var(node.output[0], empty='_'),
                              (" %s " % ops[node.op_type]).join(
            map(_rename_var, node.input)))
    name = _python_make_node_name(
        node.domain, version, node.op_type, node=True)
    attributes_str = _python_make_node_make_attribute_str(node)
    if len(node.input) > 0 and len(attributes_str) > 0:
        attributes_str = ", " + attributes_str
    output = ", ".join(map(lambda s: _rename_var(s, empty='_'), node.output))
    text = [sindent, output, " = ", name,
            '(', ', '.join(map(_rename_var, node.input)), attributes_str, ')']
    return "".join(text)


def export_template(model_onnx, templates, opset=None,  # pylint: disable=R0914
                    verbose=True, name=None,
                    rename=False, use_onnx_tensor=False,
                    autopep_options=None, function_name='create_model',
                    clean_code=True, opsets=None):
    """
    Exports an ONNX model to the onnx syntax.

    :param model_onnx: string or ONNX graph
    :param templates: exporting templates
    :param opset: opset to export to
        (None to select the one from the graph)
    :param opsets: nodes uses these opsets
    :param verbose: insert prints
    :param name: to overwrite onnx name
    :param rename: rename the names to get shorter names
    :param use_onnx_tensor: when an attribute is an array
        and its name is `'value'`, it converts that array into an
        ONNX tensor to avoid type mismatch, (operator *ConstantOfShape*, ...)
    :param autopep_options: :epkg:`autopep8` options
    :param function_name: main function name in the code
    :param clean_code: clean the code
    :return: python code
    """
    # delayed import to avoid raising an exception if not installed.
    import autopep8

    def number2name(n):
        n += 1
        seq = []
        while n >= 1:
            r = n % 26
            seq.append(r)
            n = (n - r) // 26
        return "".join(chr(65 + i) for i in reversed(seq))

    def rename_name(name, out):
        if len(name) == 0:
            if out:
                return '__'
            return "_"
        if name in dict_names:
            return dict_names[name]
        if rename:
            i = 0
            new_name = number2name(i)
            while new_name in dict_names:
                i += 1
                new_name = number2name(i)
            if len(new_name) == 0:
                raise ValueError(  # pragma: no cover
                    "Unable to rename name=%r i=%d." % (name, i))
            dict_names[name] = new_name
            dict_names[new_name] = new_name
            return new_name
        return name

    # unique_function_domain_version
    unique_function_domain_version = set()
    if hasattr(model_onnx, 'functions'):
        for f in model_onnx.functions:
            unique_function_domain_version.add((f.domain, 1))
    unique_function_domain_version = list(
        sorted(unique_function_domain_version))

    # containers
    context = {'main_model': model_onnx,
               'printable_graph': printable_graph,
               'xop_make_node_name': _xop_make_node_name,
               'python_make_node': _python_make_node,
               'python_make_node_name': _python_make_node_name,
               'unique_function_domain_version': unique_function_domain_version,
               'rename_var': _rename_var}
    used = {}

    # opset
    if hasattr(model_onnx, 'opset_import'):
        if opsets is None:
            opsets = {}
        else:
            opsets = opsets.copy()
        for oimp in model_onnx.opset_import:
            if oimp.domain == '' and opset is None:
                opsets[oimp.domain] = oimp.version
                opset = oimp.version
            else:
                opsets[oimp.domain] = opset
        context['opsets'] = opsets
        context['target_opset'] = opset
    else:
        context['opsets'] = opsets
    if opsets is None:
        raise ValueError("opsets cannot be None.")

    if hasattr(model_onnx, 'graph'):
        graph = model_onnx.graph
    else:
        graph = model_onnx
    dict_names = {}
    if rename:
        for o in graph.input:
            dict_names[o.name] = o.name
        for o in graph.output:
            dict_names[o.name] = o.name

    # inits
    unique_operators = set()
    initializers = []
    for init in graph.initializer:
        init_name = rename_name(init.name, out=False)
        value = numpy_helper.to_array(init)
        initializers.append((init_name, value))
    context['initializers'] = initializers
    context['initializers_dict'] = {k: v for k, v in initializers}

    # functions
    functions = []
    fct_dict = {}
    if hasattr(model_onnx, 'functions'):
        from ..npy.xop import OnnxOperatorFunction
        for fct in model_onnx.functions:
            used = {}
            opsets_fct = {}
            for oimp in fct.opset_import:
                if oimp.domain == '' and opset is None:
                    opsets_fct[oimp.domain] = oimp.version
                else:
                    opsets_fct[oimp.domain] = opset
            functions.append(
                (fct.domain, fct.name,
                 {'proto': fct,
                  'opsets': opsets_fct,
                  'nodes': _nodes(fct, rename_name, used, fct.output,
                                  use_onnx_tensor, templates, verbose,
                                  opset, rename, autopep_options,
                                  fct.name, [], unique_operators,
                                  opsets=opsets)}))
            if fct.name in fct_dict:
                fct_dict[fct.name].append(fct)
            else:
                fct_dict[fct.name] = [fct]
        context['OnnxOperatorFunction'] = OnnxOperatorFunction
    context['functions'] = functions
    context['functions_dict'] = fct_dict

    # inputs
    inputs = []
    for inp in graph.input:
        try:
            elem_type = get_tensor_elem_type(inp)
        except TypeError:
            # not a tensor
            inputs.append((inp.name, None, None))
            continue
        shape = get_tensor_shape(inp)
        inputs.append((inp.name, elem_type, shape))
    context['inputs'] = inputs

    # outputs
    outputs = []
    for inp in graph.output:
        try:
            elem_type = get_tensor_elem_type(inp)
        except TypeError:
            # not a tensor
            outputs.append((inp.name, None, None))
            continue
        shape = get_tensor_shape(inp)
        outputs.append((inp.name, elem_type, shape))
    context['outputs'] = outputs

    # node
    output_names = set(o.name for o in graph.output)
    subgraphs = []
    context['graph'] = graph
    context['nodes'] = _nodes(
        graph, rename_name, used, output_names, use_onnx_tensor,
        templates, verbose, opset, rename, autopep_options, name,
        subgraphs, unique_operators, opsets=opsets)

    # graph
    context['name'] = (name or graph.name).replace("(", "_").replace(")", "")
    context['function_name'] = function_name
    context['indent'] = textwrap.indent
    if hasattr(model_onnx, 'graph'):
        context['ir_version'] = model_onnx.ir_version
        context['producer_name'] = model_onnx.producer_name
        context['domain'] = model_onnx.domain
        context['model_version'] = model_onnx.model_version
        context['doc_string'] = model_onnx.doc_string
        context['metadata'] = {
            p.key: p.value for p in model_onnx.metadata_props}
    else:
        # subgraph
        context['ir_version'] = None
        context['producer_name'] = None
        context['domain'] = None
        context['model_version'] = None
        context['doc_string'] = ""
        context['metadata'] = {}

    # common context
    context['unique_operators'] = [dict(domain=o[0], name=o[1], classname=o[2])
                                   for o in sorted(unique_operators)]
    context['skip_inits'] = {}
    context['subgraphs'] = subgraphs

    mark_inits = {}

    # First rendering to detect any unused or replaced initializer.
    from jinja2 import Template  # delayed import
    template = Template(templates)
    final = template.render(
        enumerate=enumerate, sorted=sorted, len=len, map=map,
        select_attribute=select_attribute, repr=repr,
        TENSOR_TYPE_TO_NP_TYPE=TENSOR_TYPE_TO_NP_TYPE,
        make_numpy_code=lambda *args, **kwargs: make_numpy_code(
            *args, context=context, used=used, mark_inits=mark_inits,
            **kwargs),
        make_tf2onnx_code=lambda *args, **kwargs: make_tf2onnx_code(
            *args, context=context, used=used, mark_inits=mark_inits,
            **kwargs),
        verbose=verbose, **context)

    skip_inits = set()
    for k, v in mark_inits.items():
        if len(v) == len(used[k]):
            # One initializers was removed.
            skip_inits.add(k)

    if len(skip_inits) > 0:
        # Second rendering if needed when an initializer was replaced
        # or removed.
        context['skip_inits'] = skip_inits
        # Again with skip_inits.
        final = template.render(
            enumerate=enumerate, sorted=sorted, len=len,
            make_numpy_code=lambda *args, **kwargs: make_numpy_code(
                *args, context=context, used=used, mark_inits=mark_inits,
                **kwargs),
            make_tf2onnx_code=lambda *args, **kwargs: make_tf2onnx_code(
                *args, context=context, used=used, mark_inits=mark_inits,
                **kwargs),
            verbose=verbose, **context)

    final += "\n"
    if not verbose:
        rows = final.split("\n")
        final = "\n".join(_ for _ in rows if not _.endswith("#  verbose"))
    if clean_code:
        return autopep8.fix_code(final, options=autopep_options)
    return final


def export2onnx(model_onnx, opset=None, verbose=True, name=None, rename=False,
                autopep_options=None):
    """
    Exports an ONNX model to the :epkg:`onnx` syntax.

    :param model_onnx: string or ONNX graph
    :param opset: opset to export to
        (None to select the one from the graph)
    :param verbose: inserts prints
    :param name: to overwrite onnx name
    :param rename: rename the names to get shorter names
    :param autopep_options: :epkg:`autopep8` options
    :return: python code

    The following example shows what a python code creating a graph
    implementing the KMeans would look like.

    .. runpython::
        :showcode:
        :process:

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnx_tools.onnx_export import export2onnx

        X = numpy.arange(20).reshape(10, 2).astype(numpy.float32)
        tr = KMeans(n_clusters=2)
        tr.fit(X)

        onx = to_onnx(tr, X, target_opset=14)
        code = export2onnx(onx)

        print(code)
    """
    if isinstance(model_onnx, str):
        model_onnx = onnx.load(model_onnx)

    if not isinstance(model_onnx, ModelProto):
        raise TypeError(  # pragma: no cover
            f"The function expects a ModelProto not {type(model_onnx)!r}.")
    code = export_template(model_onnx, templates=get_onnx_template(),
                           opset=opset, verbose=verbose, name=name,
                           rename=rename, use_onnx_tensor=True,
                           autopep_options=autopep_options)
    return code


def export2tf2onnx(model_onnx, opset=None, verbose=True, name=None,
                   rename=False, autopep_options=None):
    """
    Exports an ONNX model to the :epkg:`tensorflow-onnx` syntax.

    :param model_onnx: string or ONNX graph
    :param opset: opset to export to
        (None to select the one from the graph)
    :param verbose: inserts prints
    :param name: to overwrite onnx name
    :param rename: rename the names to get shorter names
    :param autopep_options: :epkg:`autopep8` options
    :return: python code

    .. runpython::
        :showcode:
        :process:

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnx_tools.onnx_export import export2tf2onnx

        X = numpy.arange(20).reshape(10, 2).astype(numpy.float32)
        tr = KMeans(n_clusters=2)
        tr.fit(X)

        onx = to_onnx(tr, X, target_opset=14)
        code = export2tf2onnx(onx)

        print(code)
    """
    if isinstance(model_onnx, str):
        model_onnx = onnx.load(model_onnx)

    if not isinstance(model_onnx, ModelProto):
        raise TypeError(  # pragma: no cover
            f"The function expects a ModelProto not {type(model_onnx)!r}.")
    code = export_template(model_onnx, templates=get_tf2onnx_template(),
                           opset=opset, verbose=verbose, name=name,
                           rename=rename, use_onnx_tensor=True,
                           autopep_options=autopep_options)
    code = code.replace("], ]", "]]")
    return code


def export2numpy(model_onnx, opset=None, verbose=True, name=None,
                 rename=False, autopep_options=None):
    """
    Exports an ONNX model to the :epkg:`numpy` syntax.
    The exports does not work with all operators.

    :param model_onnx: string or ONNX graph
    :param opset: opset to export to
        (None to select the one from the graph)
    :param verbose: inserts prints
    :param name: to overwrite onnx name
    :param rename: rename the names to get shorter names
    :param autopep_options: :epkg:`autopep8` options
    :return: python code

    .. runpython::
        :showcode:
        :process:

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnx_tools.onnx_export import export2numpy

        X = numpy.arange(20).reshape(10, 2).astype(numpy.float32)
        tr = KMeans(n_clusters=2)
        tr.fit(X)

        onx = to_onnx(tr, X, target_opset=14)
        code = export2numpy(onx)

        print(code)

    This can be applied to the decomposition of an einsum
    equation into simple matrix operations.

    .. runpython::
        :showcode:
        :process:

        import numpy
        from mlprodict.testing.einsum import decompose_einsum_equation
        from mlprodict.onnx_tools.onnx_export import export2numpy

        x1 = numpy.arange(8).reshape(2, 2, 2).astype(numpy.float32)
        x2 = numpy.arange(4).reshape(2, 2).astype(numpy.float32)
        r = numpy.einsum("bac,cd->ad", x1, x2)

        seq_clean = decompose_einsum_equation(
            "bac,cd->ad", strategy='numpy', clean=True)
        onx = seq_clean.to_onnx("Y", "X1", "X2", dtype=numpy.float32)
        code = export2numpy(onx, name="einsum")
        print(code)
    """
    if isinstance(model_onnx, str):
        model_onnx = onnx.load(model_onnx)

    if not isinstance(model_onnx, ModelProto):
        raise TypeError(  # pragma: no cover
            f"The function expects a ModelProto not {type(model_onnx)!r}.")
    code = export_template(model_onnx, templates=get_numpy_template(),
                           opset=opset, verbose=verbose, name=name,
                           rename=rename, autopep_options=autopep_options)
    for i in range(-6, 6):
        code = code.replace("axis=tuple([%d])" % i, "axis=%d" % i)
        code = code.replace("tuple([%d])" % i, "(%d, )" % i)
    return code


def export2cpp(model_onnx, opset=None, verbose=True, name=None, rename=False,
               autopep_options=None):
    """
    Exports an ONNX model to the :epkg:`c` syntax.

    :param model_onnx: string or ONNX graph
    :param opset: opset to export to
        (None to select the one from the graph)
    :param verbose: inserts prints
    :param name: to overwrite onnx name
    :param rename: rename the names to get shorter names
    :param autopep_options: :epkg:`autopep8` options
    :return: python code

    The following example shows what a python code creating a graph
    implementing the KMeans would look like.

    .. runpython::
        :showcode:
        :process:

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnx_tools.onnx_export import export2cpp

        X = numpy.arange(20).reshape(10, 2).astype(numpy.float32)
        tr = KMeans(n_clusters=2)
        tr.fit(X)

        onx = to_onnx(tr, X, target_opset=14)
        code = export2cpp(onx)

        print(code)
    """
    if isinstance(model_onnx, str):
        model_onnx = onnx.load(model_onnx)

    if not isinstance(model_onnx, ModelProto):
        raise TypeError(  # pragma: no cover
            f"The function expects a ModelProto not {type(model_onnx)!r}.")
    code = export_template(model_onnx, templates=get_cpp_template(),
                           opset=opset, verbose=verbose, name=name,
                           rename=rename, use_onnx_tensor=True,
                           autopep_options=autopep_options,
                           clean_code=False)
    return code


def export2xop(model_onnx, opset=None, verbose=True, name=None, rename=False,
               autopep_options=None):
    """
    Exports an ONNX model to the XOP syntax.

    :param model_onnx: string or ONNX graph
    :param opset: opset to export to
        (None to select the one from the graph)
    :param verbose: inserts prints
    :param name: to overwrite onnx name
    :param rename: rename the names to get shorter names
    :param autopep_options: :epkg:`autopep8` options
    :return: python code

    The following example shows what a python code creating a graph
    implementing the KMeans would look like.

    .. runpython::
        :showcode:
        :process:

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnx_tools.onnx_export import export2xop

        X = numpy.arange(20).reshape(10, 2).astype(numpy.float32)
        tr = KMeans(n_clusters=2)
        tr.fit(X)

        onx = to_onnx(tr, X, target_opset=14)
        code = export2xop(onx)

        print(code)
    """
    if isinstance(model_onnx, str):
        model_onnx = onnx.load(model_onnx)

    if not isinstance(model_onnx, ModelProto):
        raise TypeError(  # pragma: no cover
            f"The function expects a ModelProto not {type(model_onnx)!r}.")
    code = export_template(model_onnx, templates=get_xop_template(),
                           opset=opset, verbose=verbose, name=name,
                           rename=rename, use_onnx_tensor=True,
                           autopep_options=autopep_options)
    return code


def export2python(model_onnx, opset=None, verbose=True, name=None, rename=False,
                  autopep_options=None, function_name='main'):
    """
    Exports an ONNX model to the *python* syntax.

    :param model_onnx: string or ONNX graph
    :param opset: opset to export to
        (None to select the one from the graph)
    :param verbose: inserts prints
    :param name: to overwrite onnx name
    :param rename: rename the names to get shorter names
    :param autopep_options: :epkg:`autopep8` options
    :param function_name: main function name
    :return: python code

    The following example shows what a python code creating a graph
    implementing the KMeans would look like.

    .. runpython::
        :showcode:
        :process:

        import numpy
        from sklearn.cluster import KMeans
        from mlprodict.onnx_conv import to_onnx
        from mlprodict.onnx_tools.onnx_export import export2python

        X = numpy.arange(20).reshape(10, 2).astype(numpy.float32)
        tr = KMeans(n_clusters=2)
        tr.fit(X)

        onx = to_onnx(tr, X, target_opset=14)
        code = export2python(onx)

        print(code)
    """
    if isinstance(model_onnx, str):
        model_onnx = onnx.load(model_onnx)

    if not isinstance(model_onnx, ModelProto):
        raise TypeError(  # pragma: no cover
            f"The function expects a ModelProto not {type(model_onnx)!r}.")
    code = export_template(model_onnx, templates=get_python_template(),
                           opset=opset, verbose=verbose, name=name,
                           rename=rename, use_onnx_tensor=True,
                           autopep_options=autopep_options,
                           clean_code=True, function_name=function_name)
    return code



"""
@file
@brief Rewrites some of the converters implemented in
:epkg:`sklearn-onnx`.
"""
import numbers
import numpy
from scipy.sparse import isspmatrix
from onnx import TensorProto
from skl2onnx.operator_converters.support_vector_machines import (
    convert_sklearn_svm_regressor)
from skl2onnx.common.data_types import guess_numpy_type, guess_proto_type
from skl2onnx.common._apply_operation import (
    apply_cast, apply_add, apply_div, apply_mul, apply_concat,
    apply_less, apply_abs)


def _op_type_domain_regressor(dtype):
    """
    Defines *op_type* and *op_domain* based on `dtype`.
    """
    if dtype == numpy.float32:
        return 'SVMRegressor', 'ai.onnx.ml', 1
    if dtype == numpy.float64:
        return 'SVMRegressorDouble', 'mlprodict', 1
    raise RuntimeError(  # pragma: no cover
        f"Unsupported dtype {dtype}.")


def _op_type_domain_classifier(dtype):
    """
    Defines *op_type* and *op_domain* based on `dtype`.
    """
    if dtype == numpy.float32:
        return 'SVMClassifier', 'ai.onnx.ml', 1
    if dtype == numpy.float64:
        return 'SVMClassifierDouble', 'mlprodict', 1
    raise RuntimeError(  # pragma: no cover
        f"Unsupported dtype {dtype}.")


def new_convert_sklearn_svm_regressor(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    op_type, op_domain, op_version = _op_type_domain_regressor(dtype)
    convert_sklearn_svm_regressor(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)


def new_convert_sklearn_svm_classifier(scope, operator, container):
    """
    Rewrites the converters implemented in
    :epkg:`sklearn-onnx` to support an operator supporting
    doubles.
    """
    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32
    op_type, op_domain, op_version = _op_type_domain_classifier(dtype)
    _convert_sklearn_svm_classifier(
        scope, operator, container, op_type=op_type, op_domain=op_domain,
        op_version=op_version)


def _convert_sklearn_svm_classifier(
        scope, operator, container,
        op_type='SVMClassifier', op_domain='ai.onnx.ml', op_version=1):
    """
    Converter for model
    `SVC <https://scikit-learn.org/stable/modules/
    generated/sklearn.svm.SVC.html>`_,
    `NuSVC <https://scikit-learn.org/stable/modules/
    generated/sklearn.svm.NuSVC.html>`_.
    The converted model in ONNX produces the same results as the
    original model except when probability=False:
    *onnxruntime* and *scikit-learn* do not return the same raw
    scores. *scikit-learn* returns aggregated scores
    as a *matrix[N, C]* coming from `_ovr_decision_function
    <https://github.com/scikit-learn/scikit-learn/blob/master/
    sklearn/utils/multiclass.py#L402>`_. *onnxruntime* returns
    the raw score from *svm* algorithm as a *matrix[N, (C(C-1)/2]*.
    """
    from sklearn.svm import NuSVC, SVC
    proto_dtype = guess_proto_type(operator.inputs[0].type)
    if proto_dtype != TensorProto.DOUBLE:  # pylint: disable=E1101
        proto_dtype = TensorProto.FLOAT  # pylint: disable=E1101
        numpy_type = numpy.float32
    else:
        numpy_type = numpy.float64

    svm_attrs = {'name': scope.get_unique_operator_name('SVMc')}
    op = operator.raw_operator
    if isinstance(op.dual_coef_, numpy.ndarray):
        coef = op.dual_coef_.ravel()
    else:
        coef = op.dual_coef_
    intercept = op.intercept_
    if isinstance(op.support_vectors_, numpy.ndarray):
        support_vectors = op.support_vectors_.ravel()
    elif isspmatrix(op.support_vectors_):
        support_vectors = op.support_vectors_.toarray().ravel()
    else:
        support_vectors = op.support_vectors_

    svm_attrs['kernel_type'] = op.kernel.upper()
    svm_attrs['kernel_params'] = [float(_)
                                  for _ in [op._gamma, op.coef0, op.degree]]
    svm_attrs['support_vectors'] = support_vectors

    if (operator.type in ['SklearnSVC', 'SklearnNuSVC'] or isinstance(
            op, (SVC, NuSVC))) and len(op.classes_) == 2:
        if isspmatrix(coef):
            coef_dense = coef.toarray().ravel()
            svm_attrs['coefficients'] = -coef_dense
        else:
            svm_attrs['coefficients'] = -coef
        svm_attrs['rho'] = -intercept
    else:
        if isspmatrix(coef):
            svm_attrs['coefficients'] = coef.todense()
        else:
            svm_attrs['coefficients'] = coef
        svm_attrs['rho'] = intercept

    handles_ovr = False
    svm_attrs['coefficients'] = svm_attrs['coefficients'].astype(numpy_type)
    svm_attrs['support_vectors'] = svm_attrs['support_vectors'].astype(
        numpy_type)
    svm_attrs['rho'] = svm_attrs['rho'].astype(numpy_type)

    options = container.get_options(op, dict(raw_scores=False))
    use_raw_scores = options['raw_scores']

    if operator.type in ['SklearnSVC', 'SklearnNuSVC'] or isinstance(
            op, (SVC, NuSVC)):
        if len(op.probA_) > 0:
            svm_attrs['prob_a'] = op.probA_.astype(numpy_type)
        else:
            handles_ovr = True
        if len(op.probB_) > 0:
            svm_attrs['prob_b'] = op.probB_.astype(numpy_type)

        if (hasattr(op, 'decision_function_shape') and
                op.decision_function_shape == 'ovr' and handles_ovr and
                len(op.classes_) > 2):
            output_name = scope.get_unique_variable_name('before_ovr')
        elif len(op.classes_) == 2 and use_raw_scores:
            output_name = scope.get_unique_variable_name('raw_scores')
        else:
            output_name = operator.outputs[1].full_name

        svm_attrs['post_transform'] = 'NONE'
        svm_attrs['vectors_per_class'] = op.n_support_.tolist()

        label_name = operator.outputs[0].full_name
        probability_tensor_name = output_name

        if all(isinstance(i, (numbers.Real, bool, numpy.bool_))
                for i in op.classes_):
            labels = [int(i) for i in op.classes_]
            svm_attrs['classlabels_ints'] = labels
        elif all(isinstance(i, str) for i in op.classes_):
            labels = [str(i) for i in op.classes_]
            svm_attrs['classlabels_strings'] = labels
        else:
            raise RuntimeError(f"Invalid class label type '{op.classes_}'.")

        svm_out = scope.get_unique_variable_name('SVM02')
        container.add_node(
            op_type, operator.inputs[0].full_name,
            [label_name, svm_out],
            op_domain=op_domain, op_version=op_version, **svm_attrs)
        apply_cast(scope, svm_out, probability_tensor_name,
                   container, to=proto_dtype)
        if len(op.classes_) == 2 and use_raw_scores:
            minus_one = scope.get_unique_variable_name('minus_one')
            container.add_initializer(minus_one, proto_dtype, [], [-1])
            container.add_node(
                'Mul', [output_name, minus_one], operator.outputs[1].full_name,
                name=scope.get_unique_operator_name('MulRawScores'))
    else:
        raise ValueError("Unknown support vector machine model type found "
                         "'{0}'.".format(operator.type))

    if (hasattr(op, 'decision_function_shape') and
            op.decision_function_shape == 'ovr' and handles_ovr and
            len(op.classes_) > 2):
        # Applies _ovr_decision_function.
        # See https://github.com/scikit-learn/scikit-learn/blob/
        # master/sklearn/utils/multiclass.py#L407:
        # ::
        #     _ovr_decision_function(dec < 0, -dec, len(self.classes_))
        #
        #     ...
        #     def _ovr_decision_function(predictions, confidences, n_classes):
        #
        #     n_samples = predictions.shape[0]
        #     votes = numpy.zeros((n_samples, n_classes))
        #     sum_of_confidences = numpy.zeros((n_samples, n_classes))
        #     k = 0
        #     for i in range(n_classes):
        #         for j in range(i + 1, n_classes):
        #             sum_of_confidences[:, i] -= confidences[:, k]
        #             sum_of_confidences[:, j] += confidences[:, k]
        #             votes[predictions[:, k] == 0, i] += 1
        #             votes[predictions[:, k] == 1, j] += 1
        #             k += 1
        #     transformed_confidences = (
        #         sum_of_confidences / (3 * (numpy.abs(sum_of_confidences) + 1)))
        #     return votes + transformed_confidences

        cst3 = scope.get_unique_variable_name('cst3')
        container.add_initializer(cst3, proto_dtype, [], [3])
        cst1 = scope.get_unique_variable_name('cst1')
        container.add_initializer(cst1, proto_dtype, [], [1])
        cst0 = scope.get_unique_variable_name('cst0')
        container.add_initializer(cst0, proto_dtype, [], [0])

        prediction = scope.get_unique_variable_name('prediction')
        if apply_less is None:
            raise RuntimeError(
                "Function apply_less is missing. "
                "onnxconverter-common is too old.")
        proto_dtype = guess_proto_type(operator.inputs[0].type)
        if proto_dtype != TensorProto.DOUBLE:  # pylint: disable=E1101
            proto_dtype = TensorProto.FLOAT  # pylint: disable=E1101
        apply_less(scope, [output_name, cst0], prediction, container)
        iprediction = scope.get_unique_variable_name('iprediction')
        apply_cast(scope, prediction, iprediction, container,
                   to=proto_dtype)

        n_classes = len(op.classes_)
        sumc_name = [scope.get_unique_variable_name('svcsumc_%d' % i)
                     for i in range(n_classes)]
        vote_name = [scope.get_unique_variable_name('svcvote_%d' % i)
                     for i in range(n_classes)]
        sumc_add = {n: [] for n in sumc_name}
        vote_add = {n: [] for n in vote_name}
        k = 0
        for i in range(n_classes):
            for j in range(i + 1, n_classes):
                name = scope.get_unique_operator_name(
                    'ArrayFeatureExtractor')
                ext = scope.get_unique_variable_name('Csvc_%d' % k)
                ind = scope.get_unique_variable_name('Cind_%d' % k)
                container.add_initializer(
                    ind, TensorProto.INT64, [], [k])  # pylint: disable=E1101
                container.add_node(
                    'ArrayFeatureExtractor', [output_name, ind],
                    ext, op_domain='ai.onnx.ml', name=name)
                sumc_add[sumc_name[i]].append(ext)

                neg = scope.get_unique_variable_name('Cneg_%d' % k)
                name = scope.get_unique_operator_name('Neg')
                container.add_node(
                    'Neg', ext, neg, op_domain='', name=name,
                    op_version=6)
                sumc_add[sumc_name[j]].append(neg)

                # votes
                name = scope.get_unique_operator_name(
                    'ArrayFeatureExtractor')
                ext = scope.get_unique_variable_name('Vsvcv_%d' % k)
                container.add_node(
                    'ArrayFeatureExtractor', [iprediction, ind],
                    ext, op_domain='ai.onnx.ml', name=name)
                vote_add[vote_name[j]].append(ext)
                neg = scope.get_unique_variable_name('Vnegv_%d' % k)
                name = scope.get_unique_operator_name('Neg')
                container.add_node(
                    'Neg', ext, neg, op_domain='', name=name,
                    op_version=6)
                neg1 = scope.get_unique_variable_name('Vnegv1_%d' % k)
                apply_add(scope, [neg, cst1], neg1, container, broadcast=1,
                          operator_name='AddCl_%d_%d' % (i, j))
                vote_add[vote_name[i]].append(neg1)

                # next
                k += 1

        for k, v in sumc_add.items():
            name = scope.get_unique_operator_name('Sum')
            container.add_node(
                'Sum', v, k, op_domain='', name=name, op_version=8)
        for k, v in vote_add.items():
            name = scope.get_unique_operator_name('Sum')
            container.add_node(
                'Sum', v, k, op_domain='', name=name, op_version=8)

        conc = scope.get_unique_variable_name('Csvcconc')
        apply_concat(scope, sumc_name, conc, container, axis=1)
        conc_vote = scope.get_unique_variable_name('Vsvcconcv')
        apply_concat(scope, vote_name, conc_vote, container, axis=1)

        conc_abs = scope.get_unique_variable_name('Cabs')
        apply_abs(scope, conc, conc_abs, container)

        conc_abs1 = scope.get_unique_variable_name('Cconc_abs1')
        apply_add(scope, [conc_abs, cst1], conc_abs1, container, broadcast=1,
                  operator_name='AddF0')
        conc_abs3 = scope.get_unique_variable_name('Cconc_abs3')
        apply_mul(scope, [conc_abs1, cst3], conc_abs3, container, broadcast=1)

        final = scope.get_unique_variable_name('Csvcfinal')
        apply_div(
            scope, [conc, conc_abs3], final, container, broadcast=0)

        output_name = operator.outputs[1].full_name
        apply_add(
            scope, [conc_vote, final], output_name, container, broadcast=0,
            operator_name='AddF1')



"""
@file
@brief Measures time processing for ONNX models.
"""
import numpy
from sklearn import __all__ as sklearn__all__, __version__ as sklearn_version
from ... import __version__ as ort_version
from .validate_helper import default_time_kwargs, measure_time


def make_n_rows(x, n, y=None):
    """
    Multiplies or reduces the rows of x to get
    exactly *n* rows.

    @param      x       matrix
    @param      n       number of rows
    @param      y       target (optional)
    @return             new matrix or two new matrices if y is not None
    """
    if n < x.shape[0]:
        if y is None:
            return x[:n].copy()
        return x[:n].copy(), y[:n].copy()
    if len(x.shape) < 2:
        r = numpy.empty((n, ), dtype=x.dtype)
        if y is not None:
            ry = numpy.empty((n, ), dtype=y.dtype)  # pragma: no cover
        for i in range(0, n, x.shape[0]):
            end = min(i + x.shape[0], n)
            r[i: end] = x[0: end - i]
            if y is not None:
                ry[i: end] = y[0: end - i]  # pragma: no cover
    else:
        r = numpy.empty((n, x.shape[1]), dtype=x.dtype)
        if y is not None:
            if len(y.shape) < 2:
                ry = numpy.empty((n, ), dtype=y.dtype)
            else:
                ry = numpy.empty((n, y.shape[1]), dtype=y.dtype)
        for i in range(0, n, x.shape[0]):
            end = min(i + x.shape[0], n)
            try:
                r[i: end, :] = x[0: end - i, :]
            except ValueError as e:  # pragma: no cover
                raise ValueError(
                    "Unexpected error: r.shape={} x.shape={} end={} i={}".format(
                        r.shape, x.shape, end, i)) from e
            if y is not None:
                if len(y.shape) < 2:
                    ry[i: end] = y[0: end - i]
                else:
                    ry[i: end, :] = y[0: end - i, :]
    if y is None:
        return r
    return r, ry


def benchmark_fct(fct, X, time_limit=4, obs=None, node_time=False,
                  time_kwargs=None, skip_long_test=True):
    """
    Benchmarks a function which takes an array
    as an input and changes the number of rows.

    @param      fct             function to benchmark, signature
                                is `fct(xo)`
    @param      X               array
    @param      time_limit      above this time, measurement is stopped
    @param      obs             all information available in a dictionary
    @param      node_time       measure time execution for each node in the graph
    @param      time_kwargs     to define a more precise way to measure a model
    @param      skip_long_test  skips tests for high values of N if they seem too long
    @return                     dictionary with the results

    The function uses *obs* to reduce the number of tries it does.
    :epkg:`sklearn:gaussian_process:GaussianProcessRegressor`
    produces huge *NxN* if predict method is called
    with ``return_cov=True``.
    The default for *time_kwargs* is the following:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        from mlprodict.onnxrt.validate.validate_helper import default_time_kwargs
        import pprint
        pprint.pprint(default_time_kwargs())

    See also notebook :ref:`onnxnodetimerst` to see how this function
    can be used to measure time spent in each node.
    """
    if time_kwargs is None:
        time_kwargs = default_time_kwargs()  # pragma: no cover

    def make(x, n):
        return make_n_rows(x, n)

    def allow(N, obs):
        if obs is None:
            return True  # pragma: no cover
        prob = obs['problem']
        if "-cov" in prob and N > 1000:
            return False  # pragma: no cover
        return True

    Ns = list(sorted(time_kwargs))
    res = {}
    for N in Ns:
        if not isinstance(N, int):
            raise RuntimeError(  # pragma: no cover
                f"time_kwargs ({type(time_kwargs)}) is wrong:\n{time_kwargs}")
        if not allow(N, obs):
            continue  # pragma: no cover
        x = make(X, N)
        number = time_kwargs[N]['number']
        repeat = time_kwargs[N]['repeat']
        if node_time:
            fct(x)
            main = None
            for __ in range(repeat):
                agg = None
                for _ in range(number):
                    ms = fct(x)[1]
                    if agg is None:
                        agg = ms
                        for row in agg:
                            row['N'] = N
                    else:
                        if len(agg) != len(ms):
                            raise RuntimeError(  # pragma: no cover
                                f"Not the same number of nodes {len(agg)} != {len(ms)}.")
                        for a, b in zip(agg, ms):
                            a['time'] += b['time']
                if main is None:
                    main = agg
                else:
                    if len(agg) != len(main):
                        raise RuntimeError(  # pragma: no cover
                            f"Not the same number of nodes {len(agg)} != {len(main)}.")
                    for a, b in zip(main, agg):
                        a['time'] += b['time']
                        a['max_time'] = max(
                            a.get('max_time', b['time']), b['time'])
                        a['min_time'] = min(
                            a.get('min_time', b['time']), b['time'])
            for row in main:
                row['repeat'] = repeat
                row['number'] = number
                row['time'] /= repeat * number
                if 'max_time' in row:
                    row['max_time'] /= number
                    row['min_time'] /= number
                else:
                    row['max_time'] = row['time']  # pragma: no cover
                    row['min_time'] = row['time']  # pragma: no cover
            res[N] = main
        else:
            res[N] = measure_time(fct, x, repeat=repeat,
                                  number=number, div_by_number=True)
        if (skip_long_test and not node_time and
                res[N] is not None and
                res[N].get('ttime', time_limit) >= time_limit):
            # too long
            break  # pragma: no cover
    if node_time:
        rows = []
        for _, v in res.items():
            rows.extend(v)
        return rows
    return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.onnx_pb import TensorProto
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from ._op import OpRun


class Cast(OpRun):

    atts = {'to': None}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Cast.atts,
                       **options)
        if self.to == TensorProto.STRING:  # pylint: disable=E1101
            self._dtype = numpy.str_
        else:
            self._dtype = TENSOR_TYPE_TO_NP_TYPE[self.to]
        self._cast = lambda x: x.astype(self._dtype)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and x.flags['WRITEABLE']:
            return self._run_inplace(x)
        return (self._cast(x), )

    def _run_inplace(self, x):
        if x.dtype == self._dtype:
            return (x, )
        return (self._cast(x), )


class CastLike(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc, **options)

    def _run(self, x, y, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.inplaces.get(0, False) and x.flags['WRITEABLE']:
            return self._run_inplace(x, y)
        return (x.astype(y.dtype), )

    def _run_inplace(self, x, y):
        if x.dtype == y.dtype:
            return (x, )
        return (x.astype(y.dtype), )



"""
@file
@brief Inspired from sklearn-onnx, handles two backends.
"""
from ...onnxrt import OnnxInference
from .utils_backend_common_compare import compare_runtime_session


class MockVariableName:
    "A string."

    def __init__(self, name):
        self.name = name

    @property
    def shape(self):
        "returns shape"
        raise NotImplementedError(  # pragma: no cover
            f"No shape for '{self.name}'.")

    @property
    def type(self):
        "returns type"
        raise NotImplementedError(  # pragma: no cover
            f"No type for '{self.name}'.")


class MockVariableNameShape(MockVariableName):
    "A string and a shape."

    def __init__(self, name, sh):
        MockVariableName.__init__(self, name)
        self._shape = sh

    @property
    def shape(self):
        "returns shape"
        return self._shape


class MockVariableNameShapeType(MockVariableNameShape):
    "A string and a shape and a type."

    def __init__(self, name, sh, stype):
        MockVariableNameShape.__init__(self, name, sh)
        self._stype = stype

    @property
    def type(self):
        "returns type"
        return self._stype


class OnnxInference2(OnnxInference):
    "onnxruntime API"

    def run(self, name, inputs, *args, **kwargs):  # pylint: disable=W0221
        "onnxruntime API"
        res = OnnxInference.run(self, inputs, **kwargs)
        if name is None:
            return [res[n] for n in self.output_names]
        if name in res:  # pragma: no cover
            return res[name]
        raise RuntimeError(  # pragma: no cover
            f"Unable to find output '{name}'.")

    def get_inputs(self):
        "onnxruntime API"
        return [MockVariableNameShapeType(*n) for n in self.input_names_shapes_types]

    def get_outputs(self):
        "onnxruntime API"
        return [MockVariableNameShape(*n) for n in self.output_names_shapes]

    def run_in_scan(self, inputs, attributes=None, verbose=0, fLOG=None):
        "Instance to run in operator scan."
        return OnnxInference.run(
            self, inputs, attributes=attributes, verbose=verbose, fLOG=fLOG)


def compare_runtime(test, decimal=5, options=None,
                    verbose=False, context=None, comparable_outputs=None,
                    intermediate_steps=False, classes=None,
                    disable_optimisation=False):
    """
    The function compares the expected output (computed with
    the model before being converted to ONNX) and the ONNX output
    produced with module :epkg:`onnxruntime` or :epkg:`mlprodict`.

    :param test: dictionary with the following keys:
        - *onnx*: onnx model (filename or object)
        - *expected*: expected output (filename pkl or object)
        - *data*: input data (filename pkl or object)
    :param decimal: precision of the comparison
    :param options: comparison options
    :param context: specifies custom operators
    :param verbose: in case of error, the function may print
        more information on the standard output
    :param comparable_outputs: compare only these outputs
    :param intermediate_steps: displays intermediate steps
        in case of an error
    :param classes: classes names (if option 'nocl' is used)
    :param disable_optimisation: disable optimisation the runtime may do
    :return: tuple (outut, lambda function to run the predictions)

    The function does not return anything but raises an error
    if the comparison failed.
    """
    return compare_runtime_session(
        OnnxInference2, test, decimal=decimal, options=options,
        verbose=verbose, context=context,
        comparable_outputs=comparable_outputs,
        intermediate_steps=intermediate_steps,
        classes=classes, disable_optimisation=disable_optimisation)



"""
@file
@brief Class ShapeResult
"""
from enum import Enum
import numpy
from .shape_excs import (
    ShapeInferenceException, NotImplementedShapeInferenceError,
    ShapeInferenceDimensionError)


class OnnxKind(Enum):
    """
    Describes a result type.
    """
    Tensor = 0
    Sequence = 1
    Map = 2


class ShapeConstraint:
    """
    One constraint.

    :param name: variable name
    :param values: set of possible values
    """

    def __init__(self, name, values):
        if name == '?':
            raise ValueError(  # pragma: no cover
                "Name cannot be '?'.")
        if not isinstance(values, set):
            raise TypeError(  # pragma: no cover
                f"values must be a set not {type(values)!r}.")
        self.name = name
        self.values = values

    def __eq__(self, other):
        "usual"
        if self.name != other.name:
            return False
        if self.values != other.values:
            return False
        return True

    def __repr__(self):
        "usual"
        return f"{self.__class__.__name__}({self.name!r}, {self.values!r})"

    def merge(self, cst):
        """
        Merges this constraint with *cst* into this one.
        """
        if isinstance(cst, list):
            for c in cst:
                self.merge(c)
            return
        self.values = self.values.intersection(cst.values)

    def copy(self, deep=False):
        """
        Makes a copy of the object.
        """
        return ShapeConstraint(self.name, self.values.copy())


class ShapeConstraintList:
    """
    A list of ShapeConstraint.
    """

    def __init__(self):
        self.csts = []

    def __contains__(self, cst):
        for a in self.csts:
            if cst == a:
                return True
        return False

    def append(self, cst):
        "Appends a new constraint to the list."
        self.csts.append(cst)

    def __repr__(self):
        return f"ShapeConstraintList({self.csts!r})"

    def __iter__(self):
        for c in self.csts:
            yield c

    def __len__(self):
        return len(self.csts)

    def copy(self, deep=False):
        """
        Copies the object.
        """
        cp = ShapeConstraintList()
        if deep:
            cp.csts = [v.copy(deep=deep) for v in self]
        else:
            cp.csts = self.csts.copy()
        return cp


class ShapeResult:
    """
    Contains information about shape and type of a result
    in an onnx graph.

    :param name: result name
    :param shape: shape if the result is a tensor
    :param dtype: element type if the result is a tensor
    :param sparse: is the tensor sparse
    :param mtype: kind of the result (see class @see cl OnnxKind)
    :param constraints: list of constraints applying on variables
    """

    def __init__(self, name, shape=None, dtype=None, sparse=False,
                 mtype=OnnxKind.Tensor, constraints=None):
        if not isinstance(name, str):
            raise TypeError(  # pragma: no cover
                f"name must be a string not {type(name)!r}.")
        if not isinstance(sparse, bool):
            raise TypeError(  # pragma: no cover
                f"sparse must be a boolean not {sparse!r}.")
        if not isinstance(mtype, OnnxKind):
            raise TypeError(  # pragma: no cover
                f"mtype must be of type OnnxKind not {type(mtype)!r}.")
        self.shape = list(shape)
        for i in range(0, len(self.shape)):  # pylint: disable=C0200
            if shape[i] in ('', None, '?'):
                raise ValueError(  # pragma: no cover
                    f"All dimensions must an int or a variable name, {shape} is not.")
        self.name = name
        self.mtype = mtype
        self.dtype = dtype
        self.sparse = sparse
        if constraints is None:
            self.constraints = ShapeConstraintList()
        elif isinstance(constraints, ShapeConstraintList):
            self.constraints = constraints
        else:
            raise TypeError(  # pragma: no cover
                "constraints must be of type(ShapeConstraintList).")

    def is_compatible(self, shape):
        """
        Tells if this shape is compatible with the given tuple.

        :param shape: tuple
        :return: boolean
        """
        if isinstance(shape, numpy.ndarray):
            shape = shape.shape
        if all(map(lambda x: isinstance(x, int), self.shape)):
            return tuple(self.shape) == tuple(shape)
        raise NotImplementedError(f"{self!r} ? {shape!r}")

    def copy(self, deep=False):
        """
        Returns a copy for the result.
        """
        return ShapeResult(self.name, self.shape, self.dtype, self.sparse,
                           self.mtype, self.constraints.copy(deep=deep))

    def __repr__(self):
        """
        Usual
        """
        if len(self.constraints) > 0:
            return "%s(%r, %r, %r, sparse=%r, mtype=%r, constraints=%r)" % (
                self.__class__.__name__, self.name, self.shape, self.dtype,
                self.sparse, self.mtype, self.constraints)
        if self.mtype != OnnxKind.Tensor:
            return "%s(%r, %r, %r, sparse=%r, mtype=%r)" % (
                self.__class__.__name__, self.name, self.shape, self.dtype,
                self.sparse, self.mtype)
        if self.sparse:
            return "%s(%r, %r, %r,sparse=%r)" % (
                self.__class__.__name__, self.name, self.shape, self.dtype,
                self.sparse)
        return "%s(%r, %r, %r)" % (
            self.__class__.__name__, self.name, self.shape, self.dtype)

    def __eq__(self, shape):
        """
        Tells if two shapes are identical.
        """
        return (self.mtype == shape.mtype and self.shape == shape.shape and
                self.dtype == shape.dtype and self.sparse == shape.sparse)

    def n_dims(self):
        """
        Returns the number of dimensions if it is a tensor.
        Raises an exception otherwise.
        """
        if self.mtype != OnnxKind.Tensor:
            raise ShapeInferenceException(  # pragma: no cover
                f"This shape is not a tensor {self!r}.")
        return len(self.shape)

    def merge(self, other_result):
        """
        Merges constraints from *other_results* into *self*.
        """
        if self.mtype != other_result.mtype:
            raise RuntimeError(  # pragma: no cover
                f"Unable to merge {self!r} and {other_result!r}.")
        if (len(self.shape) != 0 and len(other_result.shape) != 0 and
                len(self.shape) != len(other_result.shape)):
            raise ShapeInferenceDimensionError(  # pragma: no cover
                f"Length mismatch, unable to merge {self!r} and {other_result!r}.")
        updated = False
        if other_result.constraints is not None:
            for c in other_result.constraints:
                if c not in self.constraints:
                    self.constraints.append(c)
                    updated = True

        if len(self.shape) == 0 and len(other_result.shape) > 0:
            # Then self.shape is unknown and the other one is.
            self.shape = other_result.shape.copy()
            return True

        for a, b in zip(self.shape, other_result.shape):
            if a == b:
                continue
            if isinstance(a, int) and isinstance(b, int):
                raise RuntimeError(
                    f"Inconsistancy between {self!r} and {other_result!r}.")
            elif isinstance(a, str):
                c = ShapeConstraint(a, {b})
                if c not in self.constraints:
                    updated = True
                    self.constraints.append(c)
            elif isinstance(b, str):
                c = ShapeConstraint(b, {a})
                if c not in self.constraints:
                    updated = True
                    self.constraints.append(c)
            else:
                raise NotImplementedError(  # pragma: no cover
                    f"Merge not implemented between {self!r} and {other_result!r}.")
        return updated

    def resolve(self, variables):
        """
        Results variables in a shape using values stored
        in *variables*. It does not copy any constraints.

        :param variables: dictionary `{ name: values }`
        :return: new ShapeResult
        """
        res = ShapeResult(self.name, shape=self.shape, dtype=self.dtype,
                          sparse=self.sparse, mtype=self.mtype)
        for i in range(len(res.shape)):  # pylint: disable=C0200
            v = res.shape[i]
            if isinstance(v, str):
                if v in variables:
                    vals = variables[v]
                    if vals is None:
                        # size unknown
                        continue
                    if len(vals) == 1:
                        res.shape[i] = list(vals)[0]
                    else:
                        res.shape[i] = set(vals)
                else:
                    raise RuntimeError(  # pragma: no cover
                        f"Unable to resolve shape {self!r} due to missing {v!r}.")
        return res

    @staticmethod
    def broadcast(sh1, sh2, name=None, dtype=None, same_type=True):
        """
        Broadcasts dimensions for an element wise operator.

        :param sh1: ShapeResult
        :param sh2: ShapeResult
        :param name: name of the output ShapeResult
        :param dtype: type of the result or the same as the first
            element if None
        :param same_type: check the type are the same
        :return: ShapeResult
        """
        if not isinstance(sh1, ShapeResult):
            raise TypeError(  # pragma: no cover
                f"Unexpected type for sh1 {type(sh1)!r}.")
        if not isinstance(sh2, ShapeResult):
            raise TypeError(  # pragma: no cover
                f"Unexpected type for sh2 {type(sh2)!r}.")
        if sh1.mtype != OnnxKind.Tensor:
            raise TypeError(  # pragma: no cover
                f"sh1 must be a tensor not {sh1.mtype!r}.")
        if sh2.mtype != OnnxKind.Tensor:
            raise TypeError(  # pragma: no cover
                f"sh2 must be a tensor not {sh2.mtype!r}.")
        if same_type and sh1.dtype != sh2.dtype:
            raise ShapeInferenceException(  # pragma: no cover
                f"Cannot broadcast shapes {sh1!r} and {sh2!r} (dtypes).")

        # Specific cases.
        if sh1.n_dims() != sh2.n_dims():
            if sh1.n_dims() == 1 and sh1.shape[0] == 1:
                return ShapeResult(
                    name, sh2.shape, dtype or sh2.dtype, sh2.sparse, sh2.mtype)
            if sh2.n_dims() == 1 and sh2.shape[0] == 1:
                return ShapeResult(
                    name, sh1.shape, dtype or sh1.dtype, sh1.sparse, sh1.mtype)
            if sh2.n_dims() < sh1.n_dims() and sh1.shape[-sh2.n_dims():] == sh2.shape:
                return ShapeResult(
                    name, sh1.shape, dtype or sh1.dtype, sh1.sparse, sh1.mtype)
            raise NotImplementedShapeInferenceError(  # pragma: no cover
                "Broadcasting is only implemented for shape of the same "
                "size, shapes are %r and %r." % (sh1, sh2))

        # Other cases.
        constraints = ShapeConstraintList()
        shape = []
        for a, b in zip(sh1.shape, sh2.shape):
            if isinstance(a, int) and isinstance(b, int):
                if a != b:
                    if min(a, b) == 1:
                        d = max(a, b)
                    else:
                        raise ShapeInferenceException(  # pragma: no cover
                            "Cannot broadcast shapes %r and %r (dimensions)."
                            "" % (sh1, sh2))
                else:
                    d = a
            elif isinstance(a, int):
                if a != 1:
                    d = a
                    constraints.append(ShapeConstraint(b, {1, a}))
                else:
                    d = b
            elif isinstance(b, int):
                if b != 1:
                    d = b
                    constraints.append(ShapeConstraint(a, {1, b}))
                else:
                    d = a
            elif a == b:
                d = a
            elif isinstance(a, str) and isinstance(b, str):
                if a != b:
                    # Both dimensions are variables.
                    constraints.append(ShapeConstraint(a, {1, b}))
                    constraints.append(ShapeConstraint(b, {1, a}))
                d = a
            else:
                raise ShapeInferenceException(  # pragma: no cover
                    f"Cannot broadcast shapes {sh1!r} and {sh2!r}.")
            shape.append(d)
        if name in (None, ''):
            raise ValueError(  # pragma: no cover
                "name cannot be empty.")
        res = ShapeResult(name, shape, dtype or sh1.dtype, sh1.sparse or sh2.sparse,
                          sh1.mtype, constraints)
        return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Scan(OpRun):

    atts = {
        'body': None,
        'num_scan_inputs': None,
        'scan_input_axes': [],
        'scan_input_directions': [],
        'scan_output_axes': [],
        'scan_output_directions': []
    }

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Scan.atts,
                       **options)
        if not hasattr(self.body, 'run'):
            raise RuntimeError(  # pragma: no cover
                f"Parameter 'body' must have a method 'run', type {type(self.body)}.")
        self.input_directions_ = [0 if i >= len(self.scan_input_directions) else self.scan_input_directions[i]
                                  for i in range(self.num_scan_inputs)]
        max_dir_in = max(self.input_directions_)
        if max_dir_in != 0:
            raise RuntimeError(  # pragma: no cover
                "Scan is not implemented for other output input_direction than 0.")
        self.input_axes_ = [0 if i >= len(self.scan_input_axes) else self.scan_input_axes[i]
                            for i in range(self.num_scan_inputs)]
        max_axe_in = max(self.input_axes_)
        if max_axe_in != 0:
            raise RuntimeError(  # pragma: no cover
                "Scan is not implemented for other input axes than 0.")
        self.input_names = self.body.input_names
        self.output_names = self.body.output_names
        self._run_meth = (self.body.run_in_scan
                          if hasattr(self.body, 'run_in_scan')
                          else self.body.run)

    def _common_run_shape(self, *args):
        num_loop_state_vars = len(args) - self.num_scan_inputs
        num_scan_outputs = len(args) - num_loop_state_vars

        output_directions = [0 if i >= len(self.scan_output_directions) else self.scan_output_directions[i]
                             for i in range(num_scan_outputs)]
        max_dir_out = max(output_directions)
        if max_dir_out != 0:
            raise RuntimeError(  # pragma: no cover
                "Scan is not implemented for other output output_direction than 0.")
        output_axes = [0 if i >= len(self.scan_output_axes) else self.scan_output_axes[i]
                       for i in range(num_scan_outputs)]
        max_axe_out = max(output_axes)
        if max_axe_out != 0:
            raise RuntimeError(  # pragma: no cover
                "Scan is not implemented for other output axes than 0.")

        state_names_in = self.input_names[:self.num_scan_inputs]
        state_names_out = self.output_names[:len(state_names_in)]
        scan_names_in = self.input_names[num_loop_state_vars:]
        scan_names_out = self.output_names[num_loop_state_vars:]
        scan_values = args[num_loop_state_vars:]

        states = args[:num_loop_state_vars]

        return (num_loop_state_vars, num_scan_outputs, output_directions,
                max_dir_out, output_axes, max_axe_out, state_names_in,
                state_names_out, scan_names_in, scan_names_out,
                scan_values, states)

    def _run(self, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        (num_loop_state_vars, num_scan_outputs, output_directions,  # pylint: disable=W0612
         max_dir_out, output_axes, max_axe_out, state_names_in,  # pylint: disable=W0612
         state_names_out, scan_names_in, scan_names_out,  # pylint: disable=W0612
         scan_values, states) = self._common_run_shape(*args)  # pylint: disable=W0612

        max_iter = args[num_loop_state_vars].shape[self.input_axes_[0]]
        results = [[] for _ in scan_names_out]

        for iter in range(max_iter):
            inputs = {}
            for name, value in zip(state_names_in, states):
                inputs[name] = value
            for name, value in zip(scan_names_in, scan_values):
                inputs[name] = value[iter]

            try:
                outputs = self._run_meth(inputs)
            except TypeError as e:  # pragma: no cover
                raise TypeError(
                    f"Unable to call 'run' for type '{type(self.body)}'.") from e

            states = [outputs[name] for name in state_names_out]
            for i, name in enumerate(scan_names_out):
                results[i].append(numpy.expand_dims(outputs[name], axis=0))

        for res in results:
            conc = numpy.vstack(res)
            states.append(conc)
        return tuple(states)
